/* chisq_power_distrib.cpp
 *
 * This program creates a list of ap signals from the informix
 * database, then checks their raw data to see if their power
 * spectrum is relatively constant.  Astrophysical signals are
 * supposed to be wideband, and have constant power spectra
 * (or slightly sloped).
 * The program that does the checking is chisq_power_distrib.pro
 *
 * We do this as follows:
 * 1. Query the database to fetch desired list of signals
 * 2. Iterate through all signals
 * 3. Iterate through time loop, from start to stop time
 *     a. Initialize chisq_power_distrib.txt file, which 
 *      instructs IDL on how to process this batch using chisq_power_distrib.pro
 * 4. Iterate through inner loop, at most num_pulses times
 *     a. Write a chisq_script file to run chisq_power_distrib.txt
 *     b. read from chisq_power_distrib.out
 *     c. Loop through all num signals, creating or updating rfi entries for them 
 */

// #include "ap_config.h"

#ifdef _WIN32
#include "windows.h"
#endif

#include <cstdio>
#include <cstdlib>
#include <vector>
#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>

// #include <ctype.h>

// #include "setilib.h"
// #include "db_table.h"
// #include "sqlhdr.h"
// #include "schema_master.h"
// #include "seti_doppler.h"
#include "parse.h"
#include "xml_util.h"
#include "db_table.h"
#include "ap_schema.h"
// #include "astropulse.h"
#include "util.h"
#include "base64.h"
#include "ap_raw_data_db.h"
#undef read

#include <cstring>
#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#include <ctime>
#include <cmath>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

int main(int argc, char ** argv) {

char * db_name = "apb_db@sci_master_tcp";  // will change later if found in command line

long num_pulses = 100;
double start = 2454500.0;
double stop = 2454501.0;
float days_step = 1.0;
int debug = 0;
// selecting for raw_data is not null will also select from period = 0
char * where_clause ="";
char c_string[256];

for (int i=1; i<argc; i++) {
    if (!strncmp(argv[i], "-db_name", 8)) {
      db_name = argv[i] + 8;
    } else if (!strncmp(argv[i], "-start", 6)) {
      start = atof(argv[i] + 6);
    } else if (!strncmp(argv[i], "-stop", 5)) {
      stop = atof(argv[i] + 5);
    } else if (!strncmp(argv[i], "-days_step", 10)) {
      days_step = atof(argv[i] + 10); // fetch this many days worth of signals, each time through the t loop
    } else if (!strncmp(argv[i], "-num_pulses", 11)) {
      num_pulses = atol(argv[i] + 11);
    } else if (!strncmp(argv[i], "-debug", 6)) {
      debug = 1;
    } else if (!strncmp(argv[i], "-where_clause", 13)) {
      where_clause = argv[i] + 13;
    } else if (!strncmp(argv[i], "--help", 6)) {
      fprintf(stderr, "Usage: chisq_power_distrib [-where_clauseS] [-db_nameS] [-startN] [-stopN] [-days_stepN] [-num_pulsesN] [-debug] <br>\n");
      fprintf(stderr, "where_clause is concatenated before the base value<br>\n");
      fprintf(stderr, "Default values: <br>\n");
      fprintf(stderr, "where_clause: \"\"\n");
      fprintf(stderr, "base where clause: where [where_clause] [and] raw_data_len > 0 and raw_data is not null and time > [t] and time < [t + days_step] order by time <br>\n");   
      fprintf(stderr, "db_name: \"apb_db@sci_master_tcp\" <br>\n");
      fprintf(stderr, "start: 2454500.0 <br>\n");
      fprintf(stderr, "stop: 2454501.0 <br>\n");
      fprintf(stderr, "days_step: 1.0 <br>\n");
      fprintf(stderr, "num_pulses (in inner loop): 100 <br>\n");
      fprintf(stderr, "debug: not flagged <br>\n");
      exit(0);
    } else {
      fprintf(stderr, "Unrecognized argument: %s <br>\n", argv[i]);
      exit(-1);
    }
}

fprintf(stderr, "where clause: %s<br>\n", where_clause);
fprintf(stderr, "start time: %f<br>\n", start);
fprintf(stderr, "stop time: %f<br>\n", stop);
fprintf(stderr, "days_step: %f<br>\n", days_step);
fprintf(stderr, "num pulses in inner loop: %d<br>\n", num_pulses);
fprintf(stderr, "Running with db: %s<br>\n", db_name);
fprintf(stderr, "debug: %d <br>\n", debug);

// 1. Query the database to fetch desired list of signals

// set up informix and idl environment variables

setenv("INFORMIXDIR", "/disks/thumper/a/apps/informix", 1);
setenv("INFORMIXSQLHOSTS", "/disks/thumper/a/apps/informix/etc/sqlhosts", 1);
setenv("INFORMIXSERVER", "sci_master_tcp", 1);
setenv("IDL_PATH", ":~vonkorff/build/Josh_idl_files/waterfall:+~vonkorff/build/Josh_idl_files/time_range:+~/idl:+~korpela/idl:+/usr/local/rsi:+/home/korpela/NEWCVS/analysis/lib/idl_astro", 1);

int retval = db_change(db_name);
fprintf(stderr, "db_change(%s) returns %d\n", db_name, retval);

ap_signal this_ap_signal;

rfi this_rfi;
result this_result;
float fraction_blanked;
char rfi_where_clause[100];
char result_where_clause[100];
char nextline[1000];

// 2. Iterate through all signals

long num = 0; // count pulses in inner loop

long * ifx_id = new long[num_pulses];
long * ifx_resultid = new long[num_pulses];
int * ifx_dm = new int[num_pulses];
int * ifx_peak_bin = new int[num_pulses];
int * ifx_scale = new int[num_pulses];
int * scale = new int[num_pulses];
int * idl_peak_bin[num_pulses];

float * ifx_peak_power = new float[num_pulses];
float * log_prob = new float[num_pulses];
float * chisq = new float[num_pulses];
float * idl_peak_power = new float[num_pulses];

/*
ifx_dm[NUM_PULSES], ifx_peak_bin[NUM_PULSES], ifx_scale[NUM_PULSES];
int ifx_resultid[NUM_PULSES];
float ifx_peak_power[NUM_PULSES];
float log_prob[NUM_PULSES], chisq[NUM_PULSES], idl_peak_power[NUM_PULSES];
int scale[NUM_PULSES], idl_peak_bin[NUM_PULSES];
*/

int reached_last_pulse;
long totalpulses = 0;
long start_time = time(NULL);
long stop_time = time(NULL);
long sleep_time = 12;  // initial value only
long awake_time = 12;
int errval;	// returned by chisq_power_distrib.pro
char full_where_clause[1000];
char andstr[4] = "";
if(strcmp(where_clause, "")) sprintf(andstr, "and");

for(double t = start; t < stop; t += days_step) {

    reached_last_pulse = 0;  // we have not reached the last pulse yet.

    // Get some more signals.
    sprintf(full_where_clause, "where %s %s raw_data_len > 0 and raw_data is not null and time > %f and time < %f order by time", where_clause, andstr, t, t + days_step);   
    printf("Running with full_where_clause: %s <br>\n", full_where_clause);
    sql_run("set pdqpriority 10;");
    // open_query( where clause, order by, directives )
    this_ap_signal.open_query(full_where_clause, "", "{--+first_rows}");

    // 3. Iterate through outer loop.
    // Outer loop iterates through all batches of signals in the time range.
    // Inner loop processes each batch.
    while(!reached_last_pulse) {

    fprintf(stderr, "Beginning outer loop of days_step %f.  start time : %f.\n", days_step, t);

    // 3.a. Initialize chisq_power_distrib.txt file, which
    // instructs IDL on how to process this batch.
    FILE * chisq_power_distrib_txt = fopen("chisq_power_distrib.txt", "w");
    if(!chisq_power_distrib_txt) {
        printf("In chisq_power_distrib : error opening file chisq_power_distrib.txt for write.<br>\n");
        exit(-1);
    }
    fprintf(chisq_power_distrib_txt, "!PATH += \":~vonkorff/build/Josh_idl_files/stddev_test\"\n");
    fprintf(chisq_power_distrib_txt, ".run chisq_power_distrib.pro\n");

    // 4. Iterate through inner loop, at most num_pulses times
    for(num = 0; num < num_pulses; num++) {

        // sleep?
        if(time(NULL) >= start_time + awake_time) { 
            stop_time = time(NULL);
	    sleep_time = stop_time - start_time + 1;
            if(sleep_time > 1200) sleep_time = 1200;
            if(debug) printf("Sleeping for %ld seconds<br>\n", sleep_time);
	    while(time(NULL) < stop_time + sleep_time);
            start_time = time(NULL);
        }

	// If this happens, we've reached the last pulse available in the query
	// As always, we break with with num = the # of processed signals
        if(!(retval = this_ap_signal.get_next())) {
            reached_last_pulse = 1;
            break;
	}

        // ignore signals with less then min_bytes of raw data
        int min_bytes = 16382;
        // we used to also ignore signals with scale 0; 
        if(this_ap_signal.raw_data.size() < min_bytes) {
	   num--;  // don't count pulses without enough raw_data
           continue;
	}

        fprintf(stderr, "In chisq_power_distrib : getting next ap_signal : reached totalpulses = %ld<br>\n", totalpulses);

        ifx_id[num] = this_ap_signal.id;
        ifx_peak_bin[num] = this_ap_signal.peak_bin;
        ifx_scale[num] = this_ap_signal.scale;
        ifx_peak_power[num] = this_ap_signal.peak_power;
        ifx_dm[num] = this_ap_signal.dm; 
        ifx_resultid[num] = this_ap_signal.resultid.id;

        // make miniwu file
        std::stringstream os;
        os << num;
        std::string k_str = os.str(); //retrieve as a string

        std::string wufilename = "unknown_wu_" + k_str;
        if(Astropulse::ap_signal_to_miniwufile(this_ap_signal, wufilename, 1, "")) {
            printf("In chisq_power_distrib : error in ap_signal_to_miniwufile.<br>\n");
            exit(-1);
        }

        totalpulses++;
        fprintf(chisq_power_distrib_txt, "chisq_power_distrib, file=\"unknown_wu_%d_0.mwu\", dm=%d, scale=%d, errval\n", num, this_ap_signal.dm, this_ap_signal.scale);

    }
    fprintf(chisq_power_distrib_txt, "chisq_power_distrib_print_err, \"chisq_power_distrib_err.txt\", errval\n");

    fclose(chisq_power_distrib_txt);

    // If no pulses were found on this iteration, we shouldn't try to open the
    // chisq_power_distrib.out file, b/c it doesn't exist.
    if(num == 0) break;

    // 4. a. Write a chisq_script file to run chisq_power_distrib.txt
    FILE * chisq_script = fopen("chisq_script", "w");

    if(!chisq_script) {
        fprintf(stderr, "In chisq_power_distrib : can't open chisq_script for write.<br>\n");
        exit(-1);
    }
 
    fprintf(chisq_script, "#! /bin/csh\n");
    fprintf(chisq_script, "# Script will create a chisq power distrib output file\n");
    fprintf(chisq_script, "/usr/local/bin/idl < chisq_power_distrib.txt\n");

    fclose(chisq_script);

    if(system("chmod u+x chisq_script")) {
        fprintf(stderr, "In chisq_power_distrib : error when running chmod u+x chisq_script<br>\n");
        exit(-1);
    }

    if(system("chisq_script")) {
        fprintf(stderr, "In chisq_power_distrib : error when running chisq_script<br>\n");
        exit(-1);
    }

    FILE * chisq_power_distrib_err_txt = fopen("chisq_power_distrib_err.txt", "r");

    if(!chisq_power_distrib_err_txt) {
        printf("In chisq_power_distrib : unable to open file chisq_power_distrib_err.txt for read<br>\n");
        exit(-1);
    }

    if(fscanf(chisq_power_distrib_err_txt, "error %d", &errval) != 1) {
        printf("In chisq_power_distrib : unable to read chisq_power_distrib_err.txt<br>\n");
        exit(-1);
    }

    fclose(chisq_power_distrib_err_txt);

    if(errval != 0) {  
        printf("In chisq_power_distrib : from chisq_power_distrib.pro : errval %d<br>\n", errval);
        printf("time = %f<br>\n", t);
        printf("totalpulses = %ld<br>\n", totalpulses);
        exit(-1);
    }

    // check that we had at least num successes, then erase the log file.
    // if we didn't have at least num successes, exit with error.
    // This is actually a second way of testing for errors
    // so I should probably remove it ... but it can't hurt to leave it for now.
    
    FILE * chisq_power_distrib_log = fopen("chisq_power_distrib.log", "r");
    if(!chisq_power_distrib_log) {
        fprintf(stderr, "In chisq_power_distrib : error : cannot open chisq_power_distrib.log for read<br>\n");
        exit(-1);
    }

    for(long i = 0; i < num; i++) {
        if(fgets(nextline, 1000, chisq_power_distrib_log) == NULL) {
            fprintf(stderr, "In chisq_power_distrib : error : end of chisq_power_distrib.log file reached.<br>\n");
            fprintf(stderr, "This probably means that chisq_power_distrib.pro has crashed<br>\n");
            exit(-1);
        }
    }

    fclose(chisq_power_distrib_log);

    if(system("rm chisq_power_distrib.log")) {
        fprintf(stderr, "In chisq_power_distrib : error when running rm chisq_power_distrib.log<br>\n");
        exit(-1);
    }

    // 4. b. read from chisq_power_distrib.out
    FILE * chisq_power_distrib_out = fopen("chisq_power_distrib.out", "r");
    if(!chisq_power_distrib_out) {
        fprintf(stderr, "In chisq_power_distrib : error : cannot open chisq_power_distrib.out for read<br>\n");
        exit(-1);
    }
 
    if(debug) printf("==================================\n"); 
    for(long i = 0; i < num; i++) { 
        fscanf(chisq_power_distrib_out, "%f", &log_prob[i]);
        fscanf(chisq_power_distrib_out, "%d", &scale[i]);
        fscanf(chisq_power_distrib_out, "%f", &chisq[i]);
        fscanf(chisq_power_distrib_out, "%d", &idl_peak_bin[i]);
        fscanf(chisq_power_distrib_out, "%f", &idl_peak_power[i]);

    // all these printfs should go away when we're running for real.
        if(debug) {
            printf("i: %d\n", i);

            printf("ifx ap_signal id: %ld\n", ifx_id[i]);
            printf("ifx ap_signal resultid: %d\n", ifx_resultid[i]);
            printf("ifx ap_signal peak_bin: %d\n", ifx_peak_bin[i]);
            printf("ifx ap_signal scale: %d\n", ifx_scale[i]);
            printf("ifx ap_signal peak_power: %f\n", ifx_peak_power[i]);
            printf("ifx ap_signal dm: %d\n", ifx_dm[i]);
        
            printf("log_prob: %f\n", log_prob[i]);
            printf("chisq: %f\n", chisq[i]);
            printf("idl peak_bin of max: %d\n", idl_peak_bin[i]);
            printf("idl peak_power of max: %f\n\n", idl_peak_power[i]);
            printf("----------------------------------\n"); 
        }
    }

    fclose(chisq_power_distrib_out);

    // chisq_power_distrib.pro will append to this file, so we have to remove it
    // before the next loop, or it will just keep growing. 
    if(system("rm -f chisq_power_distrib.out")) {
        printf("In chisq_power_distrib : error : cannot rm chisq_power_distrib.out");
        exit(-1);
    }

    // 4. c. Loop through all num signals, creating or updating rfi entries for them 
    for(long i = 0; i < num; i++) {
    
	// Access the result table to update fraction_blanked
	sprintf(result_where_clause, "where id = %d", ifx_resultid[i]);
	this_result.open_query(result_where_clause, "", "{--+first_rows}");
	retval = this_result.get_next();
        if(retval)
		fraction_blanked = this_result.fraction_blanked;	
	else    // this should never happen ... ?
		fraction_blanked = -1;
	this_result.close_query();
      
	// Access rfi table to fetch existing rfi information 
        sprintf(rfi_where_clause, "where ap_signalid = %ld", ifx_id[i]);
        this_rfi.open_query(rfi_where_clause, "", "{--+first_rows}");
        retval = this_rfi.get_next();

        // Set some members of this_rfi.
        // (whether or not it exists in the database.)
        // For chi squared:
        this_rfi.chisq = chisq[i];
        this_rfi.log_prob = log_prob[i];
        this_rfi.chisq_time = time(NULL);
        // For fraction_blanked:
	this_rfi.fraction_blanked = fraction_blanked;
        // For edge pulse:
        this_rfi.edge_pulse_diff = ifx_peak_power[i] - idl_peak_power[i];
        this_rfi.edge_pulse_time = this_rfi.chisq_time;  // = time(NULL)

        if(retval) {  // Update old rfi
            this_rfi.update();
	    if(debug) {
                printf("Updated this_rfi with id: %d\n", this_rfi.id);
                printf("ap_signalid: %d\n", this_rfi.ap_signalid.id);
                printf("chisq_time: %f\n", this_rfi.chisq_time);
	    }
        } else {  // Create new rfi
            this_rfi.ap_signalid.id = ifx_id[i];
            this_rfi.user_marked = -2;
            this_rfi.user_marked_time = 0;
	    this_rfi.rep_at_dm = -2;
            this_rfi.rep_at_dm_time = 0;
            // already set chisq, log_prob, and chisq_time 
            this_rfi.multi_beams = -2;
            this_rfi.multi_beams_time = 0;
            // already set fraction_blanked
            this_rfi.multi_pols = -2;
            this_rfi.multi_pols_time = 0;
            this_rfi.pm_dm_pair = -2;
            this_rfi.pm_dm_pair_time = 0;
            // already set edge_pulse_diff

            retval = this_rfi.insert();
            printf("got this retval from this_rfi.insert: %d\n", retval);
	    if(debug) {
                printf("Inserted this_rfi with id: %d\n", this_rfi.id);
	        // ap_signal is a composite type, and requires ap_signalid.id
	        printf("ap_signalid: %d\n", this_rfi.ap_signalid.id);
                printf("fraction_blanked: %f\n", this_rfi.fraction_blanked);
                printf("chisq_time: %f\n", this_rfi.chisq_time);
	    }
        }

        this_rfi.close_query();
    } // inner loop

    } // outer loop

    this_ap_signal.close_query();
 
} // time loop

} // main
