/* mwu_test.cpp
 *
 * This program performs the following steps repeatedly:
 * 1. Randomly determine a power with distribution S^(-5/2)
 * 2. Call Adam's mworkunit.pro to create a mini workunit
 * 3. Call chisq_power_distrib.pro to test this mwu
 * 4. Repeat (2) with same power
 * 5. Repeat (3)
 * 6. Analyze results
 *
 * Final output:
 * 1. # of pairs for which both exceeded threshold, neither, and one only
 * 2. # of pulses with various log_prob intervals.
 */

// #include "ap_config.h"

#ifdef _WIN32
#include "windows.h"
#endif

#include <cstdio>
#include <cstdlib>
#include <vector>
#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>

// #include <ctype.h>

// #include "setilib.h"
// #include "db_table.h"
// #include "sqlhdr.h"
// #include "schema_master.h"
// #include "seti_doppler.h"
#include "parse.h"
#include "xml_util.h"
#include "db_table.h"
#include "ap_schema.h"
// #include "astropulse.h"
#include "util.h"
#include "base64.h"
#include "ap_raw_data_db.h"
#undef read

#include <cstring>
#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#include <ctime>
#include <cmath>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

int main(int argc, char ** argv) {

int debug = 0;
long numiter = 10000;

for (int i=1; i<argc; i++) {
    if (!strncmp(argv[i], "-numiter", 8)) {
        first = atol(argv[i] + 8);
    } else if (!strncmp(argv[i], "-debug", 6)) {
        debug = 1;
    } else if (!strncmp(argv[i], "--help", 6)) {
        fprintf(stderr, "Usage: mwu_test [-iterN] [-debug] <br>\n");
        fprintf(stderr, "Default values: <br>\n");
        fprintf(stderr, "iter: 10000 <br>\n");
        fprintf(stderr, "debug: not flagged <br>\n");
        exit(0);
    }
}

fprintf(stderr, "numiter: %ld<br>\n", numiter);
fprintf(stderr, "debug: %d <br>\n", debug);

// set up idl environment variables

setenv("IDL_PATH", ":+~/idl:+~korpela/idl:+/usr/local/rsi:+/home/korpela/NEWCVS/analysis/lib/idl_astro", 1);

char nextline[1000];

float * log_prob = new float[2];
float * chisq = new float[2];
float * scale = new float[2];
float * idl_peak_power = new float[2];
float * idl_peak_bin = new float[2];
int errval;	// returned by chisq_power_distrib.pro

float thresh[10] = {29.107864, 31.568300, 37.925224, 49.470177, 61.368561,
                    86.990051, 128.857971, 212.633087, 361.960083, 648.259888};

for(long iter = 0; iter < numiter; iter++) {

    // Determine power here
    float power = 30;

    for(int i = 0; i < 2; i++) {

        FILE * mwu_test_txt = fopen("mwu_test.txt", "w");
        if(!mwu_test_txt) {
            printf("In chisq_power_distrib : error opening file mwu_test.txt for write.<br>\n");
            exit(-1);
        }

        fprintf(mwu_test_txt, "!PATH += \":~vonkorff/build/Josh_idl_files/stddev_test\"\n");
        fprintf(mwu_test_txt, ".run chisq_power_distrib.pro\n");
        fprintf(mwu_test_txt, "mworkunit, file=\"mwu_test_1.mwu\", dm=%d, scale=%d\n", dm, scale);
        fprintf(mwu_test_txt, "chisq_power_distrib, file=\"mwu_test_1.mwu\", dm=%d, scale=%d, outfile=\"mwu_test.out\", logfile=\"mwu_test.log\", errval\n", dm, scale);
        fprintf(mwu_test_txt, "chisq_power_distrib_print_err, \"mwu_test_1_err.txt\", errval\n");
        fclose(mwu_test_txt);

        // 4. a. Write a mwu_script file to run mwu_test.txt
        FILE * mwu_script = fopen("mwu_script", "w");
    
        fprintf(mwu_script, "#! /bin/csh\n");
        fprintf(mwu_script, "# Script will create a mwu test output file\n");
        fprintf(mwu_script, "/usr/local/bin/idl < mwu_test.txt\n");

        fclose(mwu_script);

        if(system("chmod u+x mwu_script")) {
            fprintf(stderr, "In mwu_test : error when running chmod u+x mwu_script<br>\n");
        }

        if(system("mwu_script")) {
            fprintf(stderr, "In mwu_test_distrib : error when running mwu_script<br>\n");
            exit(-1);
        }
    
        FILE * mwu_test_err_txt = fopen("mwu_test_err.txt", "r");
    
        if(!mwu_test_err_txt) {
            printf("In mwu_test : unable to open file mwu_test_err.txt<br>\n");
            exit(-1);
        }

        if(fscanf(mwu_test_err_txt, "error %d", &errval) != 1) {
            printf("In mwu_test : unable to read mwu_test_err.txt<br>\n");
            exit(-1);
        }

        fclose(chisq_power_distrib_err_txt);

        if(errval != 0) {  
            printf("In chisq_power_distrib : from chisq_power_distrib.pro : errval %d<br>\n", errval);
            printf("outer_loops = %ld<br>\n", outer_loops);
            printf("totalpulses = %ld<br>\n", totalpulses);
            exit(-1);
        }
    }

    // check that we had at least num successes, then erase the log file.
    // if we didn't have at least num successes, exit with error.
    // This is actually a second way of testing for errors
    // so I should probably remove it ... but it can't hurt to leave it for now.
    
    FILE * mwu_test_log = fopen("mwu_test.log", "r");
    if(!mwu_test_log) {
        fprintf(stderr, "In mwu_test : error : cannot open mwu_test.log for read<br>\n");
        exit(-1);
    }

    for(int i = 0; i < 2; i++) {

        if(fgets(nextline, 1000, mwu_test_log) == NULL) {
            fprintf(stderr, "In mwu_test : error : end of mwu_test.log file reached.<br>\n");
            fprintf(stderr, "This probably means that chisq_power_distrib.pro has crashed<br>\n");
            exit(-1);
        }
    }

    fclose(mwu_test_log);

    if(system("rm mwu_test.log")) {
        fprintf(stderr, "In mwu_test : error when running rm mwu_test.log<br>\n");
        exit(-1);
    }

    // read from chisq_power_distrib.out
    FILE * mwu_test_out = fopen("mwu_test.out", "r");
    if(!mwu_test_out) {
        fprintf(stderr, "In mwu_test : error : cannot open mwu_test.out for read<br>\n");
        exit(-1);
    }
 
    if(debug) printf("==================================\n"); 
    for(long i = 0; i < num; i++) { 
        fscanf(mwu_test_out, "%f", &log_prob[i]);
        fscanf(mwu_test_out, "%d", &scale[i]);
        fscanf(mwu_test_out, "%f", &chisq[i]);
        fscanf(mwu_test_out, "%d", &idl_peak_bin[i]);
        fscanf(mwu_test_out, "%f", &idl_peak_power[i]);

    // all these printfs should go away when we're running for real.
        if(debug) {
            printf("i: %d\n", i);

            printf("log_prob: %f\n", log_prob[i]);
            printf("chisq: %f\n", chisq[i]);
            printf("idl peak_bin of max: %d\n", idl_peak_bin[i]);
            printf("idl peak_power of max: %f\n\n", idl_peak_power[i]);
            printf("----------------------------------\n"); 
        }
    }

    fclose(mwu_test_out);

    // chisq_power_distrib.pro will append to this file, so we have to remove it
    // before the next loop, or it will just keep growing. 
    if(system("rm -f mwu_test.out")) {
        printf("In mwu_test : error : cannot rm mwu_test.out");
        exit(-1);
    }

} // for loop

} // main
