/* pulsestream_mark_rfi.cpp
 *
 * This program detects 3 kinds of rfi.  It will:
 * 0. Iterate over all time ranges (1 hour in length?  1 day?)
 * 1. write a .sql for a particular time range, using C
 * 2. from C, run the .sql file using dbaccess from the command line
 * 3. This creates a text file called pulsestream.txt, listing all ap_signals within
 *    a certain time range ... perhaps at least 100,000 signals.
 * 4. from C, write an idl command file that calls a certain function
 * 5. this function should 
 *    a. call rdfloat to read in pulsestream.txt
 *    b. run rfi_rej3, possibly iteratively
 *    c. print out a NEW file, using printf
 * 6. read in pulsestream.txt
 * 7. modify the rfi rows associated with each pulse, as appropriate
 */

// #include "ap_config.h"

#ifdef _WIN32
#include "windows.h"
#endif

#include <cstdio>
#include <cstdlib>
#include <vector>
#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>

// #include <ctype.h>

// #include "setilib.h"
// #include "db_table.h"
// #include "sqlhdr.h"
// #include "schema_master.h"
// #include "seti_doppler.h"
#include "parse.h"
#include "xml_util.h"
#include "db_table.h"
#include "ap_schema.h"
// #include "astropulse.h"
#include "util.h"
#include "base64.h"
#include "ap_raw_data_db.h"
#undef read

#include <cstring>
#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#include <ctime>
#include <cmath>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

int main(int argc, char ** argv) {

char * db_name = "apb_db@sci_master_tcp";  // will change later if found in command line

int first=0;
int help = 0;
// int have_where_clause = 0;
float days_step = 1.0; // number of days to step as we iterate through the database in time order
char c_string[256];
double start = -1.0;
double stop = -1.0;

for (int i=1; i<argc; i++) {
    if (!strncmp(argv[i], "-db_name", 8)) {
        db_name = argv[i] + 8;   	// use this database
    } else if (!strncmp(argv[i], "-first", 6)) {
        first = atoi(argv[i] + 6);	// only insert/update this many pulses
    } else if (!strncmp(argv[i], "-start", 6)) {
        start = atof(argv[i] + 6);	// start time julian date
    } else if (!strncmp(argv[i], "-stop", 5)) {	
        stop = atof(argv[i] + 5);	// stop time julian date
    } else if (!strncmp(argv[i], "--help", 6)) {
	help = 1;
    } else if (!strncmp(argv[i], "-days_step", 10)) {
        days_step = atof(argv[i] + 10);	// fetch this many days worth of signals, each time through the t loop
    } else {
        fprintf(stderr, "Unrecognized argument: %s <br>\n", argv[i]);
        exit(-1);
    }
}

if(stop < 0 || start < 0 || help) {
  fprintf(stderr, "In pulsestream_mark_rfi : You must specify a -start and a -stop julian date. <br>\n");
  fprintf(stderr, "Usage: pulsestream_mark_rfi -startN -stopN [-days_step] [-db_nameNAME] [-firstN] <br>\n");
  fprintf(stderr, "Default values: <br>\n");
  fprintf(stderr, "days_step: 1.0 days <br>\n");
  fprintf(stderr, "db_name: apb_db@sci_master_tcp <br>\n");
  fprintf(stderr, "first: 0 (meaning to insert all signals) <br>\n");
  exit(-1);
}

fprintf(stderr, "Running with db: %s<br>\n", db_name);
fprintf(stderr, "Inserting first %d signals<br>\n", first);
fprintf(stderr, "Starting at julian date %f<br>\n", start);
fprintf(stderr, "Stopping at julian date %f<br>\n", stop);
fprintf(stderr, "Stepping by %f days <br>\n", days_step);
// fprintf(stderr, "Running with where_clause: %s <br>\n", where_clause.c_str());

setenv("INFORMIXDIR", "/disks/thumper/a/apps/informix", 1);
setenv("INFORMIXSQLHOSTS", "/disks/thumper/a/apps/informix/etc/sqlhosts", 1);
setenv("INFORMIXSERVER", "sci_master_tcp", 1);
setenv("IDL_PATH", ":+~/idl:+~korpela/idl:+/usr/local/rsi:+/home/korpela/NEWCVS/analysis/lib/idl_astro", 1);

int retval = db_change(db_name);	// Select a database
fprintf(stderr, "db_change(%s) returns %d\n", db_name, retval);

int id, isrfi_rep_at_dm, isrfi_multi_pols, isrfi_multi_beams;

rfi this_rfi;			// We'll select an rfi into this variable, to be updated (or inserted)
char rfi_where_clause[100];  	// Will contain a clause like "where ap_signalid = %d"
int count_pulses = 0;   	// Count the # of pulses we have inserted.  Will be limited by first
int errval;			// returned by rfi_rej3.pro

long start_time;
long stop_time;
long awake_time = 12;
long sleep_time;

// 0. Iterate over all time ranges (1 hour in length?  1 day?)

for(double t = start; t < stop; t += days_step) {

    // 1. write a .sql for a particular time range, using C
    fprintf(stderr, "Writing sql file getpulses.sql for range starting with %f, current start time %f, days step %f\n", start, t, days_step);

    FILE * getpulses_sql = fopen("getpulses.sql", "w");
  
    fprintf(getpulses_sql,  "database %s;\n", db_name);
    fprintf(getpulses_sql, "set pdqpriority 10;\n");
    fprintf(getpulses_sql, "unload to \"pulsestream.txt\" delimiter \" \"\n");
    fprintf(getpulses_sql, "select { --+first_rows }\n");
    fprintf(getpulses_sql, "ap_signal.id,dm,time,period,beam,polarization,workunit.id,peak_bin,scale,ra,decl\n");
    fprintf(getpulses_sql, "from tape,workunit,result,ap_signal where\n");
    fprintf(getpulses_sql, "period = 0                               and\n");
    fprintf(getpulses_sql, "result.versionid>=503                     and\n");
    fprintf(getpulses_sql, "result.id = ap_signal.resultid            and\n");
    fprintf(getpulses_sql, "workunit.id = result.wuid                 and\n");
    fprintf(getpulses_sql, "ap_signal.time >= %f			and\n", t);
    fprintf(getpulses_sql, "ap_signal.time < %f			and\n", t + days_step);
    fprintf(getpulses_sql, "tape.id = workunit.tape_info		\n");

    fclose(getpulses_sql);

    // 2. from C, run the .sql file using dbaccess from the command line
    fprintf(stderr, "Running sql file getpulses.sql using dbaccess from the command line.\n");

    if(system("/disks/thumper/a/apps/informix/bin/dbaccess -e < getpulses.sql")) {
        printf("In pulsestream_mark_rfi : error when running dbaccess\n");
        exit(-1);
    }

    // 3. Now we have created a text file called pulsestream.txt, listing all ap_signals within
    //    a certain time range ... perhaps up to 100,000 signals.
    //    However, we must check if there are any signals, some days have none.

    FILE * pulsestream_txt = fopen("pulsestream.txt", "r");

    if(!pulsestream_txt) {
        printf("In pulsestream_mark_rfi : error : cannot open pulsestream.txt for read.<br>\n");
        printf("This might mean that the ap_db unload command, which was supposed to create the file, failed.<br>\n");
        exit(-1);
    }
   
    fseek(pulsestream_txt, 0L, SEEK_END);
    long sz = ftell(pulsestream_txt);   
  
    if(sz == 0) {
        printf("In pulsestream_mark_rfi : day is empty, t = %f.<br>\n", t);
        fclose(pulsestream_txt);
        continue;
    }

    fclose(pulsestream_txt);
 
    // 4. from C, write an idl command file that calls a certain set of functions
    fprintf(stderr, "Writing idl command file pulsestream_mark_rfi_script\n");  
 
    FILE * pulsestream_mark_rfi_script = fopen("pulsestream_mark_rfi_script", "w");

    if(!pulsestream_mark_rfi_script) {
        printf("In pulsestream_mark_rfi : error opening file pulsestream_mark_rfi_script for write.<br>\n");
        exit(-1);
    }
    fprintf(pulsestream_mark_rfi_script, "#! /bin/csh\n");
    fprintf(pulsestream_mark_rfi_script, "# Script will create a pulsestream_with_marks.txt output file\n");
    fprintf(pulsestream_mark_rfi_script, "/usr/local/bin/idl < pulsestream_mark_rfi_idl.txt\n");
  
    fclose(pulsestream_mark_rfi_script);
   
    fprintf(stderr, "Writing idl command file pulsestream_mark_rfi_idl.txt\n");
 
    FILE * pulsestream_mark_rfi_idl_txt = fopen("pulsestream_mark_rfi_idl.txt", "w");

    if(!pulsestream_mark_rfi_idl_txt) {
        printf("In pulsestream_mark_rfi : error opening file pulsestream_mark_rfi_idl.txt for write.<br>\n");
        exit(-1);
    }
    fprintf(pulsestream_mark_rfi_idl_txt, "!PATH += \":~vonkorff/build/Josh_idl_files/rfi_rej3\"\n");

    // 5. these functions should

    //    a. call rdfloat to read in pulsestream.txt
    fprintf(pulsestream_mark_rfi_idl_txt, "rdfloat, \"pulsestream.txt\", id, dm, time, period, beam, pol, wuid, peak_bin, scale, ra, decl, /double\n");
    //    b. run rfi_rej3, 3 times.  Goal is to catch about 99% of the errors.
    fprintf(pulsestream_mark_rfi_idl_txt, ".run rfi_rej3.pro\n");
    fprintf(pulsestream_mark_rfi_idl_txt, ".run rfi_rej3_print_marked.pro\n");
    fprintf(pulsestream_mark_rfi_idl_txt, "rfi_rej3, id, dm, time, period, beam, pol, wuid, peak_bin, scale, ra, decl, isrfi_rep_at_dm, isrfi_multi_pols, isrfi_multi_beams, errval\n");
    fprintf(pulsestream_mark_rfi_idl_txt, "rfi_rej3, id, dm, time, period, beam, pol, wuid, peak_bin, scale, ra, decl, isrfi_rep_at_dm, isrfi_multi_pols, isrfi_multi_beams, errval\n");
    fprintf(pulsestream_mark_rfi_idl_txt, "rfi_rej3, id, dm, time, period, beam, pol, wuid, peak_bin, scale, ra, decl, isrfi_rep_at_dm, isrfi_multi_pols, isrfi_multi_beams, errval\n");
    //    c. print out a NEW file, using printf
    fprintf(pulsestream_mark_rfi_idl_txt, "rfi_rej3_print_marked, \"pulsestream_with_marks.txt\", \"rfi_rej3_err.txt\", id, dm, time, period, beam, pol, wuid, peak_bin, scale, ra, decl, isrfi_rep_at_dm, isrfi_multi_pols, isrfi_multi_beams, errval");

    fclose(pulsestream_mark_rfi_idl_txt);

    if(system("chmod u+x pulsestream_mark_rfi_script")) {
        printf("In pulsestream_mark_rfi : error when running chmod u+x pulsestream_mark_rfi_script<br>\n");
        exit(-1);
    }

    if(system("pulsestream_mark_rfi_script")) {
        printf("In pulsestream_mark_rfi : error when running pulsestream_mark_rfi_script<br>\n");
        exit(-1);
    }
    
    FILE * rfi_rej3_err_txt = fopen("rfi_rej3_err.txt", "r");

    if(!rfi_rej3_err_txt) {
        printf("In pulsestream_mark_rfi : unable to open file rfi_rej3_err.txt<br>\n");
        exit(-1);
    }

    if(fscanf(rfi_rej3_err_txt, "error %d", &errval) != 1) {
	printf("In pulsestream_mark_rfi : unable to read rfi_rej3_err.txt<br>\n");
        exit(-1);
    }

    fclose(rfi_rej3_err_txt); 
    
    if(errval != 0) { 
	printf("In pulsestream_mark_rfi : from rfi_rej3.pro : errval %d<br>\n", errval);
	// printf("Possible errvals include:<br>\n");
        // printf("error 1: not enough points<br>\n");
	printf("julian date: t = %f<br>\n", t);
        // if(errval == 1) continue;
        exit(-1);
    }

    //  6. read in pulsestream_with_marks.txt
    
    FILE * pulsestream_with_marks_txt = fopen("pulsestream_with_marks.txt", "r");
    if(!pulsestream_with_marks_txt) {
        printf("In pulsestream_mark_rfi : error opening file pulsestream_with_marks.txt for read.<br>\n");
        exit(-1);
    }
  
    start_time = time(NULL);
    stop_time = time(NULL);
    sleep_time = 12;  // initial value only
 
    // id, dm, time, period, beam, pol, wuid, peak_bin, scale, ra, decl,
    // isrfi_rep_at_dm, isrfi_multi_pols, isrfi_multi_beams
    while(fscanf(pulsestream_with_marks_txt, "%d %*d %*f %*f %*d %*d %*d %*ld %*d %*f %*f %d %d %d",
		&id, &isrfi_rep_at_dm, &isrfi_multi_pols, &isrfi_multi_beams) == 4) {  // want to read 4 of the 14 cols

        if(time(NULL) >= start_time + awake_time) { 
            stop_time = time(NULL);
	    sleep_time = stop_time - start_time + 1;
            if(sleep_time > 1200) sleep_time = 1200;
            printf("Sleeping for %ld seconds<br>\n", sleep_time);
            while(time(NULL) < stop_time + sleep_time);
            start_time = time(NULL);
        }

        //  7. modify the rfi rows associated with each pulse, as appropriate

        // Access rfi table to fetch existing rfi information
        sprintf(rfi_where_clause, "where ap_signalid = %d", id);
        this_rfi.open_query(rfi_where_clause, "", "{--+first_rows}");
        retval = this_rfi.get_next();

        // Set some members of this_rfi.
        // (whether or not it exists in the database.)
        this_rfi.rep_at_dm = isrfi_rep_at_dm;
        this_rfi.rep_at_dm_time = time(NULL);
        this_rfi.multi_pols = isrfi_multi_pols;
        this_rfi.multi_pols_time = time(NULL);
        this_rfi.multi_beams = isrfi_multi_beams;
        this_rfi.multi_beams_time = time(NULL);

        if(retval) {  // Update old rfi
            this_rfi.update();
            if(count_pulses % 1000 == 0) {
                printf("Updated rfi with id: %d\n", this_rfi.id);
                printf("ap_signalid: %d\n", this_rfi.ap_signalid.id);
	    }
        } else {  // Create new rfi
            this_rfi.ap_signalid = id;
            this_rfi.user_marked = -2;
            this_rfi.user_marked_time = 0;
            // already set rep_at_dm
            this_rfi.chisq = -2;
            this_rfi.log_prob = 2;
            this_rfi.chisq_time = 0;
	    // already set multi_beams
            this_rfi.fraction_blanked = -2;
	    // already set multi_pols
            this_rfi.pm_dm_pair = -2;
            this_rfi.pm_dm_pair_time = 0;
            this_rfi.edge_pulse_diff = -2;
            this_rfi.edge_pulse_time = 0;

            this_rfi.insert();
            if(count_pulses % 1000 == 0) {
                printf("Inserted this_rfi with id: %d\n", this_rfi.id);
                // ap_signal is a composite type, and requires ap_signalid.id
                printf("ap_signalid: %d\n", this_rfi.ap_signalid.id);
	    }
        }

        this_rfi.close_query();
        if(first != 0 && count_pulses++ > first) exit(0);
    }  // end while fscanf
    fclose(pulsestream_with_marks_txt);
}

}
